#![no_main]
#![no_std]

extern crate cortex_m;
#[macro_use(entry, exception)]
extern crate cortex_m_rt as rt;
extern crate panic_semihosting;

extern crate stm32f4xx_hal as hal;
// #[macro_use(block)]
// extern crate nb;

#[rtic::app(device = stm32f4xx_hal::pac, peripherals = true)]
mod app {
    use core::cell::RefCell;

    use cortex_m::interrupt::{free, Mutex};
    use cortex_m::peripheral::NVIC;
    use hal::block;
    use hal::pac::{self, USART1};
    use hal::serial::{Config, Event, Serial};
    use hal::spi::{Mode, Phase, Polarity, Spi};
    use rf4463::config::RADIO_CONFIG_500_2;
    use rf4463::Rf4463;
    use ringbuf::{Consumer, Producer, Rb, StaticRb};

    use crate::hal::{interrupt, prelude::*};

    enum SlaveCmd {
        BufferStatus,
        SendPacket,
    }

    impl SlaveCmd {
        pub fn from_u8(i: u8) -> Option<Self> {
            match i {
                0 => Some(Self::BufferStatus),
                1 => Some(Self::SendPacket),
                _ => None,
            }
        }
    }

    enum SlaveState {
        Idle,
        RecvPacket(usize),
    }

    #[shared]
    struct Shared {}

    #[local]
    struct Local {
        usart: Serial<USART1>,
        state: SlaveState,
        producer: Producer<u8, StaticRb<u8, 16384>>,
        consumer: Consumer<u8, StaticRb<u8, 16384>>,
    }

    static USART: Mutex<RefCell<Option<Serial<USART1, u8>>>> = Mutex::new(RefCell::new(None));

    static SLAVE_STATE: Mutex<RefCell<SlaveState>> = Mutex::new(RefCell::new(SlaveState::Idle));

    static BUF: Mutex<RefCell<Option<StaticRb<u8, 16384>>>> = Mutex::new(RefCell::new(None));

    const MODE: Mode = Mode {
        polarity: Polarity::IdleLow,
        phase: Phase::CaptureOnFirstTransition,
    };

    const PACKET_LEN: usize = 256 + 2 + 65;

    #[init]
    fn init(mut ctx: init::Context) -> (Shared, Local, init::Monotonics) {
        let rcc = ctx.device.RCC.constrain();

        let gpioa = ctx.device.GPIOA.split();
        let gpiob = ctx.device.GPIOB.split();

        // setup clocks
        let clocks = rcc.cfgr.use_hse(25.MHz()).sysclk(48.MHz()).freeze();

        // setup 4463 spi
        let mosi = gpioa.pa7.into_alternate();
        let miso = gpioa.pa6.into_alternate();
        let sclk = gpioa.pa5.into_alternate();

        let sdn = gpiob.pb1.into_push_pull_output();

        let cs = gpiob.pb0.into_push_pull_output();

        let spi = Spi::new(ctx.device.SPI1, (sclk, miso, mosi), MODE, 10.MHz(), &clocks);

        let delay = ctx.device.TIM5.delay_us(&clocks);
        let mut radio = Rf4463::new(spi, sdn, cs, delay, &mut RADIO_CONFIG_500_2.clone()).unwrap();

        // setup Pi UART
        let tx = gpioa.pa9;
        let rx = gpioa.pa10;

        let mut serial = Serial::new(
            ctx.device.USART1,
            (tx, rx),
            Config::default().baudrate(921_600.bps()),
            &clocks,
        )
        .unwrap();

        serial.listen(Event::Rxne);

        // enable usart interrupt
        unsafe {
            NVIC::unmask(interrupt::USART1);
        }

        let mut data = [0; PACKET_LEN];
        let mut i = 0;

        loop {
            if i == PACKET_LEN {
                radio.tx(&mut data).unwrap();
                i = 0;
                continue;
            }

            free(|cs| {
                let mut buf = BUF.borrow(cs).borrow_mut();
                let buf = buf.as_mut().unwrap();

                if let Some(b) = buf.pop() {
                    data[i] = b;

                    i += 1;

                    if buf.free_len() == 16000 {
                        //block!(USART.borrow(cs).borrow_mut().as_mut().unwrap().write(b'E')).unwrap();
                    }
                }
            });
        }
    }

    /*#[exception]
    unsafe fn HardFault(ef: &ExceptionFrame) -> ! {
        panic!("{:#?}", ef);
    }*/

    #[interrupt]
    fn USART1() {
        free(|cs| {
            let mut state = SLAVE_STATE.borrow(cs).borrow_mut();

            let mut usart = USART.borrow(cs).borrow_mut();
            let usart = usart.as_mut().unwrap();

            let mut buf = BUF.borrow(cs).borrow_mut();
            let buf = buf.as_mut().unwrap();

            let x = block!(usart.read()).unwrap();

            match *state {
                SlaveState::Idle => {
                    if let Some(cmd) = SlaveCmd::from_u8(x) {
                        match cmd {
                            SlaveCmd::BufferStatus => {
                                if buf.free_len() < PACKET_LEN {
                                    block!(usart.write(0x01)).unwrap(); // not enough space for a new packet
                                } else {
                                    block!(usart.write(0x02)).unwrap(); // safe to send a packet
                                }
                            }

                            SlaveCmd::SendPacket => {
                                *state = SlaveState::RecvPacket(PACKET_LEN - 1);
                            }
                        }
                    }
                }

                SlaveState::RecvPacket(i) => {
                    *state = if i == 0 {
                        SlaveState::Idle
                    } else {
                        SlaveState::RecvPacket(i - 1)
                    };

                    buf.push(x).unwrap();
                }
            };
        });
    }
}
